import json

from client.common.decorators import log
from client.common.errors import IncorrectDataRecivedError
from client.common.variables import MAX_MESSAGE_LENGTH, ENCODING


@log
def get_message(client):
    """
    Утилита приёма и декодирования сообщения принимает байты выдаёт словарь,
    если приняточто-то другое отдаёт ошибку значения
    """
    encoded_reply = client.recv(MAX_MESSAGE_LENGTH)  # берем из сокета байты
    if isinstance(encoded_reply, bytes):
        json_reply = encoded_reply.decode(ENCODING)  # делаем декодирование
        reply = json.loads(json_reply)  # получаем словарь
        if isinstance(reply, dict):  # проверка явл ли словарем
            return reply
        else:
            raise IncorrectDataRecivedError

    else:
        raise IncorrectDataRecivedError


@log
def send_message(sock, message):
    """
    Утилита кодирования и отправки сообщения
    принимает словарь и отправляет его
    """
    if not isinstance(message, dict):
        raise NonDictInputError
    json_message = json.dumps(message)  # из словаря message делаем json-строку
    encoded_message = json_message.encode(ENCODING)  # делаем байты
    sock.send(encoded_message)  # байты кладем в сокет



from PyQt5.QtCore import *
from PyQt5.QtWidgets import *


class UiMainClientWindow(object):
    def setup_ui(self, main_window):
        if not main_window.objectName():
            main_window.setObjectName(u"MainWindow")
        main_window.resize(813, 637)
        main_window.setStyleSheet(u"background-color: rgb(228, 231, 255);")
        self.action = QAction(main_window)
        self.action.setObjectName(u"action")
        self.menu_exit = QAction(main_window)
        self.menu_exit.setObjectName(u"action_2")
        self.menu_add_contact = QAction(main_window)
        self.menu_add_contact.setObjectName(u"action_3")
        self.menu_del_contact = QAction(main_window)
        self.menu_del_contact.setObjectName(u"action_4")
        self.centralwidget = QWidget(main_window)
        self.centralwidget.setObjectName(u"centralwidget")
        self.list_contacts = QListView(self.centralwidget)
        self.list_contacts.setObjectName(u"listView")
        self.list_contacts.setGeometry(QRect(640, 20, 131, 471))
        self.list_contacts.setStyleSheet(u"background-color: rgb(228, 231,"
                                         u" 255);")
        self.label = QLabel(self.centralwidget)
        self.label.setObjectName(u"label")
        self.label.setGeometry(QRect(640, 0, 121, 16))
        self.btn_add_contact = QPushButton(self.centralwidget)
        self.btn_add_contact.setObjectName(u"pushButton")
        self.btn_add_contact.setGeometry(QRect(640, 500, 131, 26))
        self.btn_remove_contact = QPushButton(self.centralwidget)
        self.btn_remove_contact.setObjectName(u"pushButton_2")
        self.btn_remove_contact.setGeometry(QRect(640, 540, 131, 26))
        self.list_messages = QListView(self.centralwidget)
        self.list_messages.setObjectName(u"listView_2")
        self.list_messages.setGeometry(QRect(30, 20, 581, 211))
        self.label_2 = QLabel(self.centralwidget)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setGeometry(QRect(30, 0, 211, 20))
        self.label_new_message = QLabel(self.centralwidget)
        self.label_new_message.setObjectName(u"label_3")
        self.label_new_message.setGeometry(QRect(30, 240, 600, 16))
        self.text_message = QPlainTextEdit(self.centralwidget)
        self.text_message.setObjectName(u"text_message")
        self.text_message.setGeometry(QRect(30, 260, 581, 231))
        self.btn_clear = QPushButton(self.centralwidget)
        self.btn_clear.setObjectName(u"pushButton_3")
        self.btn_clear.setGeometry(QRect(30, 540, 121, 26))
        self.btn_send = QPushButton(self.centralwidget)
        self.btn_send.setObjectName(u"pushButton_4")
        self.btn_send.setGeometry(QRect(450, 540, 161, 26))
        main_window.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(main_window)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 813, 21))
        self.menu = QMenu(self.menubar)
        self.menu.setObjectName(u"menu")
        self.menu_2 = QMenu(self.menubar)
        self.menu_2.setObjectName(u"menu_2")
        main_window.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(main_window)
        self.statusbar.setObjectName(u"statusbar")
        main_window.setStatusBar(self.statusbar)
        self.toolBar = QToolBar(main_window)
        self.toolBar.setObjectName(u"toolBar")
        main_window.addToolBar(Qt.TopToolBarArea, self.toolBar)

        self.menubar.addAction(self.menu.menuAction())
        self.menubar.addAction(self.menu_2.menuAction())
        self.menu.addAction(self.menu_exit)
        self.menu_2.addAction(self.menu_add_contact)
        self.menu_2.addAction(self.menu_del_contact)

        self.retranslate_ui(main_window)

        QMetaObject.connectSlotsByName(main_window)

    # setupUi

    def retranslate_ui(self, main_window):
        main_window.setWindowTitle(QCoreApplication.translate("MainWindow",
                                                             u"MainWindow", None))
        self.action.setText(QCoreApplication.translate("MainWindow",
                                                       u"\u0412\u044b\u0445\u043e\u0434", None))
        self.menu_exit.setText(QCoreApplication.translate("MainWindow",
                                                          u"\u0412\u044b\u0445\u043e\u0434", None))
        self.menu_add_contact.setText(QCoreApplication.translate("MainWindow",
                                                                 u"\u0414\u043e\u0431\u0430\u0432\u0438"
                                                                 u"\u0442\u044c",
                                                                 None))
        self.menu_del_contact.setText(QCoreApplication.translate("MainWindow",
                                                                 u"\u0423\u0434\u0430\u043b\u0438\0442\u044c", None))
        self.label.setText(QCoreApplication.translate("MainWindow",
                                                      u"\u0421\u043f\u0438\u0441\u043e\u043a "
                                                      u"\u043a\u043e\u043d\u0442\u0430\u043a\u0442\u043e\u0432:",
                                                      None))
        self.btn_add_contact.setText(QCoreApplication.translate("MainWindow", u"\u0414\u043e\u0431\u0430\u0432\u0438"
                                                                              u"\u0442\u044c "
                                                                              u"\u043a\u043e\u043d\u0442\u0430\u043a"
                                                                              u"\u0442", None))
        self.btn_remove_contact.setText(QCoreApplication.translate("MainWindow", u"\u0423\u0434\u0430\u043b\u0438"
                                                                                 u"\u0442\u044c "
                                                                                 u"\u043a\u043e\u043d\u0442\u0430"
                                                                                 u"\u043a\u0442", None))
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"\u0418\u0441\u0442\u043e\u0440\u0438\u044f "
                                                                      u"\u0441\u043e\u043e\u0431\u0449\u0435\u043d"
                                                                      u"\u0438\u0439:", None))
        self.label_new_message.setText(QCoreApplication.translate("MainWindow", u"\u0412\u0432\u0435\u0434\u0438"
                                                                                u"\u0442\u0435 "
                                                                                u"\u0441\u043e\u043e\u0431\u0449"
                                                                                u"\u0435\u043d\u0438\u0435:", None))
        self.btn_clear.setText(QCoreApplication.translate("MainWindow", u"\u041e\u0447\u0438\u0441\u0442\u0438\u0442"
                                                                        u"\u044c \u043f\u043e\u043b\u0435", None))
        self.btn_send.setText(QCoreApplication.translate("MainWindow", u"\u041e\u0442\u043f\u0440\u0430\u0432\u0438"
                                                                       u"\u0442\u044c "
                                                                       u"\u0441\u043e\u043e\u0431\u0449\u0435\u043d"
                                                                       u"\u0438\u0435", None))
        self.menu.setTitle(QCoreApplication.translate("MainWindow", u"\u0424\u0430\u0439\u043b", None))
        self.menu_2.setTitle(QCoreApplication.translate("MainWindow", u"\u041a\u043e\u043d\u0442\u0430\u043a\u0442"
                                                                      u"\u044b", None))
        self.toolBar.setWindowTitle(QCoreApplication.translate("MainWindow", u"toolBar", None))
    # retranslateUi



import sys
import binascii
import hashlib
import hmac
import socket
import threading
import time

from PyQt5.QtCore import pyqtSignal, QObject

from client.common.utils import *
from client.common.variables import *
from client.common.errors import ServerError

sys.path.append('../../')

# Логер и объект блокировки для работы с сокетом.
client_logger = logging.getLogger('client')
socket_lock = threading.Lock()


# Класс - Транспорт, отвечает за взаимодействие с сервером
class ClientTransport(threading.Thread, QObject):
    # Сигналы новое сообщение и потеря соединения
    new_message = pyqtSignal(dict)
    message_205 = pyqtSignal()
    connection_lost = pyqtSignal()

    def __init__(self, port, ip_address, database, username, passwd, keys):
        # Вызываем конструктор предков
        threading.Thread.__init__(self)
        QObject.__init__(self)

        # Класс База данных - работа с базой
        self.database = database
        # Имя пользователя
        self.username = username
        # Сокет для работы с сервером
        self.transport = None
        self.keys = keys
        self.password = passwd
        # Устанавливаем соединение:
        self.connection_init(port, ip_address)
        # Обновляем таблицы известных пользователей и контактов
        try:
            self.user_list_update()
            self.contacts_list_update()
        except OSError as err:
            if err.errno:
                client_logger.critical(f'Потеряно соединение с сервером.')
                raise ServerError('Потеряно соединение с сервером!')
            client_logger.error('вышло время во время соединения при '
                                'обновлении списков пользователей.')
        except json.JSONDecodeError:
            client_logger.critical(f'Потеряно соединение с сервером.')
            raise ServerError('Потеряно соединение с сервером!')
            # Флаг продолжения работы транспорта.
        self.running = True

    def connection_init(self, port, ip):
        '''Метод отвечающий за установку соединения с сервером'''
        # Инициализация сокета и сообщение серверу о нашем появлении
        self.transport = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Таймаут необходим для освобождения сокета.
        self.transport.settimeout(5)

        # Соединяемся, 5 попыток соединения, флаг успеха ставим в True,
        # если удалось
        connected = False
        for i in range(5):
            client_logger.info(f'Попытка подключения №{i + 1}')
            try:
                self.transport.connect((ip, port))
            except (OSError, ConnectionRefusedError):
                pass
            else:
                connected = True
                client_logger.debug("Соединение установлено.")
                break
            time.sleep(1)

        # Если соединится не удалось - исключение
        if not connected:
            client_logger.critical('Не удалось установить соединение'
                                   ' с сервером')
            raise ServerError('Не удалось установить соединение '
                              'с сервером')

        client_logger.debug('Starting auth dialog.')

        # Запускаем процедуру авторизации
        # Получаем хэш пароля
        passwd_bytes = self.password.encode('utf-8')
        salt = self.username.lower().encode('utf-8')
        passwd_hash = hashlib.pbkdf2_hmac('sha512', passwd_bytes,
                                          salt, 10000)
        passwd_hash_string = binascii.hexlify(passwd_hash)

        client_logger.debug(f'Password hash ready:'
                            f' {passwd_hash_string}')

        # Получаем публичный ключ и декодируем его из байтов
        pubkey = self.keys.publickey().export_key().decode('ascii')

        # Авторизируемся на сервере
        with socket_lock:
            presense = {
                ACTION: PRESENCE,
                TIME: time.time(),
                USER: {
                    ACCOUNT_NAME: self.username,
                    PUBLIC_KEY: pubkey
                }
            }
            client_logger.debug(f"Сообщение о присутствии = {presense}")
            # Отправляем серверу приветственное сообщение.
            try:
                send_message(self.transport, presense)
                ans = get_message(self.transport)
                client_logger.debug(f'Server response = {ans}.')
                # Если сервер вернул ошибку, бросаем исключение.
                if REPLY in ans:
                    if ans[REPLY] == 400:
                        raise ServerError(ans[ERROR])
                    elif ans[REPLY] == 511:
                        # Если всё нормально, то продолжаем процедуру
                        # авторизации.
                        ans_data = ans[DATA]
                        hash = hmac.new(passwd_hash_string,
                                        ans_data.encode('utf-8'), 'MD5')
                        digest = hash.digest()
                        my_ans = REPLY_511
                        my_ans[DATA] = binascii.b2a_base64(
                            digest).decode('ascii')
                        send_message(self.transport, my_ans)
                        self.process_server_ans(get_message(self.transport))
            except (OSError, json.JSONDecodeError) as err:
                client_logger.debug(f'Connection error.', exc_info=err)
                raise ServerError('Сбой соединения в процессе авторизации.')

    # Функция обрабатывающяя сообщения от сервера. Ничего не возращает.
    # Генерирует исключение при ошибке.
    def process_server_ans(self, message):
        client_logger.debug(f'Разбор сообщения от сервера: {message}')

        # Если это подтверждение чего-либо
        if REPLY in message:
            if message[REPLY] == 200:
                return
            elif message[REPLY] == 400:
                raise ServerError(f'{message[ERROR]}')
            elif message[REPLY] == 205:
                self.user_list_update()
                self.contacts_list_update()
                self.message_205.emit()
            else:
                client_logger.error(f'Принят неизвестный код подтверждения'
                                    f' {message[REPLY]}')

        # Если это сообщение от пользователя добавляем в базу,
        # даём сигнал о новом сообщении
        elif ACTION in message and message[ACTION] == MESSAGE and \
                SENDER in message and DESTINATION in message \
                and MESSAGE_TEXT in message and \
                message[DESTINATION] == self.username:
            client_logger.debug(f'Получено сообщение от пользователя '
                                f'{message[SENDER]}:{message[MESSAGE_TEXT]}')

            self.new_message.emit(message)

    def contacts_list_update(self):
        '''Метод обновляющий с сервера список контактов.'''
        self.database.contacts_clear()
        client_logger.debug(f'Запрос контакт листа для пользователся '
                            f'{self.name}')
        req = {
            ACTION: GET_CONTACTS,
            TIME: time.time(),
            USER: self.username
        }
        client_logger.debug(f'Сформирован запрос {req}')
        with socket_lock:
            send_message(self.transport, req)
            ans = get_message(self.transport)
        client_logger.debug(f'Получен ответ {ans}')
        if REPLY in ans and ans[REPLY] == 202:
            for contact in ans[LIST_INFO]:
                self.database.add_contact(contact)
        else:
            client_logger.error('Не удалось обновить список контактов.')

    def user_list_update(self):
        '''Метод обновления таблицы известных пользователей.'''
        client_logger.debug(f'Запрос списка известных пользователей '
                            f'{self.username}')
        req = {
            ACTION: USERS_REQUEST,
            TIME: time.time(),
            ACCOUNT_NAME: self.username
        }
        with socket_lock:
            send_message(self.transport, req)
            ans = get_message(self.transport)
        if REPLY in ans and ans[REPLY] == 202:
            self.database.add_users(ans[LIST_INFO])
        else:
            client_logger.error('Не удалось обновить список '
                                'известных пользователей.')

    def key_request(self, user):
        '''Метод, запрашивающий с сервера публичный ключ пользователя.'''
        client_logger.debug(f'Запрос публичного ключа для {user}')
        req = {
            ACTION: PUBLIC_KEY_REQUEST,
            TIME: time.time(),
            ACCOUNT_NAME: user
        }
        with socket_lock:
            send_message(self.transport, req)
            ans = get_message(self.transport)
        if REPLY in ans and ans[REPLY] == 511:
            return ans[DATA]
        else:
            client_logger.error(f'Не удалось получить ключ собеседника{user}.')

    def add_contact(self, contact):
        client_logger.debug(f'Создание контакта {contact}')
        req = {
            ACTION: ADD_CONTACT,
            TIME: time.time(),
            USER: self.username,
            ACCOUNT_NAME: contact
        }
        with socket_lock:
            send_message(self.transport, req)
            self.process_server_ans(get_message(self.transport))

    def remove_contact(self, contact):
        '''Метод удаления клиента на сервере.'''
        client_logger.debug(f'Удаление контакта {contact}')
        req = {
            ACTION: REMOVE_CONTACT,
            TIME: time.time(),
            USER: self.username,
            ACCOUNT_NAME: contact
        }
        with socket_lock:
            send_message(self.transport, req)
            self.process_server_ans(get_message(self.transport))

    def transport_shutdown(self):
        '''Метод закрытия соединения, отправляет сообщение о выходе.'''
        self.running = False
        message = {
            ACTION: EXIT,
            TIME: time.time(),
            ACCOUNT_NAME: self.username
        }
        with socket_lock:
            try:
                send_message(self.transport, message)
            except OSError:
                pass
        client_logger.debug('Транспорт завершает работу.')
        time.sleep(0.5)

    def send_message(self, to, message):
        '''Метод отправки сообщения на сервер.'''
        message_dict = {
            ACTION: MESSAGE,
            SENDER: self.username,
            DESTINATION: to,
            TIME: time.time(),
            MESSAGE_TEXT: message
        }
        client_logger.debug(f'Сформирован словарь сообщения: {message_dict}')

        # Необходимо дождаться освобождения сокета для отправки сообщения
        with socket_lock:
            send_message(self.transport, message_dict)
            self.process_server_ans(get_message(self.transport))
            client_logger.info(f'Отправлено сообщение для пользователя {to}')

    def run(self):
        '''Метод, содержащий основной цикл работы транспортного потока.'''
        client_logger.debug('Запущен процесс - приёмник сообщений с сервера.')
        while self.running:
            # Отдыхаем секунду и снова пробуем захватить сокет.
            # если не сделать здесь задержку, то отправка может достаточно
            # долго ждать освобождения сокета.
            time.sleep(1)
            message = None
            with socket_lock:
                try:
                    self.transport.settimeout(0.5)
                    message = get_message(self.transport)
                except OSError as err:
                    if err.errno:
                        client_logger.critical(f'Потеряно соединение '
                                               f'с сервером.')
                        self.running = False
                        self.connection_lost.emit()
                # Проблемы с соединением
                except (ConnectionError, ConnectionAbortedError,
                        ConnectionResetError,
                        json.JSONDecodeError, TypeError):
                    client_logger.debug(f'Потеряно соединение с сервером.')
                    self.running = False
                    self.connection_lost.emit()
                finally:
                    self.transport.settimeout(5)

                # Если сообщение получено, то вызываем функцию обработчик:
                if message:
                    client_logger.debug(f'Принято сообщение с сервера:'
                                        f' {message}')
                    self.process_server_ans(message)



import dis


# Метакласс для проверки соответствия сервера:
class ServerVerifier(type):
    def __init__(self, clsname, bases, clsdict):

        methods = []

        attrs = []

        for func in clsdict:

            try:

                return_func = dis.get_instructions(clsdict[func])

            except TypeError:
                pass
            else:

                for i in return_func:
                    print(i)

                    if i.opname == 'LOAD_GLOBAL':  # i - Instruction,
                        if i.argval not in methods:
                            methods.append(i.argval)
                    elif i.opname == 'LOAD_ATTR':
                        if i.argval not in attrs:
                            attrs.append(i.argval)
        print(methods)

        if 'connect' in methods:
            raise TypeError('Использование метода connect недопустимо'
                            ' в серверном классе')

        if not ('SOCK_STREAM' in attrs and 'AF_INET' in attrs):
            raise TypeError('Некорректная инициализация сокета.')

        super().__init__(clsname, bases, clsdict)


class ClientVerifier(type):
    def __init__(self, clsname, bases, clsdict):
        # Список методов, которые используются в функциях класса:
        methods = []
        for func in clsdict:
            # Пробуем
            try:
                return_func = dis.get_instructions(clsdict[func])
                # Если не функция то ловим исключение
            except TypeError:
                pass
            else:
                # Раз функция, разбираем код, получая используемые методы.
                for i in return_func:
                    if i.opname == 'LOAD_GLOBAL':
                        if i.argval not in methods:
                            methods.append(i.argval)

        for command in ('accept', 'listen', 'socket'):
            if command in methods:
                raise TypeError('В классе обнаружено использование '
                                'запрещённого метода')

        if 'get_message' in methods or 'send_message' in methods:
            pass
        else:
            raise TypeError('Отсутствуют вызовы функций, '
                            'работающих с сокетами.')
        super().__init__(clsname, bases, clsdict)



import sys
import os
import logging
from logging import INFO

sys.path.append('../')

# создаём формировщик логов (formatter):
client_formatter = logging.Formatter('%(asctime)s %(levelname)s %(filename)s %(message)s')

# Подготовка имени файла для логирования
path = os.getcwd()
path = os.path.join(path, '../logs/client.log')

# создаём потоки вывода логов
stream_handler = logging.StreamHandler(sys.stderr)
stream_handler.setFormatter(client_formatter)
stream_handler.setLevel(logging.ERROR)
log_file = logging.FileHandler(path, encoding='utf8')
log_file.setFormatter(client_formatter)

# создаём регистратор и настраиваем его
logger = logging.getLogger('client')
logger.addHandler(stream_handler)
logger.addHandler(log_file)
logger.setLevel(logging.DEBUG)

# отладка
if __name__ == '__main__':
    logger.critical('Критическая ошибка')
    logger.error('Ошибка')
    logger.debug('Отладочная информация')
    logger.info('Информационное сообщение')



import sys

import logging


server_logger = logging.getLogger('server')


class Port:
    def __set__(self, instance, value):
        if not 1023 < value < 65536:
            server_logger.critical(
                f'Попытка запуска с указанием неподходящего порта '
                f'{value}. Допустимы адреса с 1024 до 65535.')
            sys.exit(1)
            # Если порт прошел проверку, добавляем его в список атрибутов
            # экземпляра

        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        # owner -   объект класса Server
        # name - port
        self.name = name



from setuptools import setup, find_packages

setup(name="mess_client_anna",
      version="0.0.1",
      description="mess_client_anna",
      author="Anna M.",
      author_email="an.malch@gmail.com",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex']
      )